#Simulate the genotypes of 5SNPs for 200 subjects, assign the first 100 to cohort A and the rest to cohort B.
geno=rep(NA, 10)
for (i in 1:5){
    allele=sample(c("A", "T", "C", "G"),2)
    add=sample(c(0,1,2), size=200, replace=T)
    genotypes=sapply(add, function(x){
   ifelse(x==0, return(rep(allele[1], 2)), ifelse(x==2, return(rep(allele[2], 2)), return(c(allele[1], allele[2]))));
    })
    
       geno=cbind(geno, t(genotypes))
}
geno=geno[,-1]


###Ex1: haplotype association test for a population-based cohort A
geno_A=data.frame(id=1:100, geno[1:100,])
lab=paste("SNP", 1:5, sep="")


source("zaykinfam.r")

#simulate the phenotype and the covariate age:
pheno=data.frame(id=1:100, FG=runif(100, 4.3,20), age=runif(100, 22,70))
res.cohort_A=zaykinfam(FG~-1+age, geno_A,  lab, id="id", lab, pheno, fullkins=NULL, cutoff=0)

###Ex2: haplotype association test for a family-based cohort B
#Simulate the genotypes of 5SNPs for 100 unrelated subjects
id=1:100
geno_B=data.frame(id, geno[101:200,])
lab=paste("SNP", 1:5, sep="")

#simulate the pedigree structure:
famid=rep(1:25, each=4)
ped=data.frame(famid, id, fa=rep(0, 100), mo=rep(0,100))

for (i in unique(famid)){
    tmp=subset(ped, famid==i)
    tmp[3:4,-c(1:2)]=rep(tmp$id[1:2], each=2)
    ped[ped$famid==i,]=tmp
}

library(kinship2)
kmat=makekinship(ped$famid, ped$id, ped$fa, ped$mo)

source("zaykinfam.r")

#simulate the phenotype and the covariate age:
pheno=data.frame(id=1:100, FG=runif(100, 4.3,20), age=runif(100, 22,70))
res.cohort_B=zaykinfam(FG~-1+age, geno_B,  lab, id="id", lab, pheno, fullkins=kmat, cutoff=0)

###Meta-analyzing the results of cohort A and cohort B:
N_cohort=2
N_variant=5
zaykin_all=list()
zaykin_all[[1]]=res.cohort_A
zaykin_all[[2]]=res.cohort_B

#Define a pre-processing function meta_hap_prep:
meta_hap_prep<-function(z_obj, cohort, annote=TRUE){
    happrob3=z_obj$haplofreq;

    if (annote==TRUE){
        names(happrob3)[ncol(happrob3)]=paste(cohort, ".hapcat", sep="")
    }


    beta=z_obj$estimate;
    if (class(beta)=="matrix"){
        beta=beta[,"Estimate"]
    }
    
    
    var=z_obj$variance;
    colnames(var)=names(beta);
    rownames(var)=names(beta);
    #var=var[!is.na(beta), !is.na(beta)]
    beta=beta[!is.na(beta)]
    beta2=beta[grep("hap", names(beta), value=T)];
    var2=var[names(beta2), names(beta2)]
    if (annote==TRUE){
        happrob3=subset(happrob3, eval(parse(text=paste(cohort, ".hapcat", sep=""))) %in% substr(names(beta2), 4, nchar(names(beta2))))
    } else {
        happrob3=subset(happrob3, hapcat %in% substr(names(beta2), 4, nchar(names(beta2))))
    }        
    RES=list(Hap.prob=happrob3, Beta=beta2, Variance=var2);
    return(RES);
}

#using the pre-processing function to QC:
cohort_exclude=NULL
for (m in 1:N_cohort){
    #File.tmp=load(File);
    File.load=zaykin_all[[m]];
    cohort=paste("cohort", m, sep="");
    File.prep<-meta_hap_prep(File.load, cohort, annote=TRUE);
    
    
    sigma_i=solve(File.prep$Variance, tol=10^(-50))
    if (any(eigen(sigma_i)$values<0)){
        cohort_exclude=c(cohort_exclude, m)
    }
}
cohort_list=1:N_cohort
cohort_list=cohort_list[!(cohort_list %in% cohort_exclude)]

#Meta-anayze the cohorts passing QC
cohort1_FG.tmp<-meta_hap_prep(zaykin_all[[cohort_list[1]]], paste("cohort",cohort_list[1], sep="") , annote=FALSE)
merge.tmp=cohort1_FG.tmp$Hap.prob[,c(1:N_variant, (N_variant+2))]
merge.tmp[, paste("cohort", cohort_list[1], ".hapcat", sep="")]=merge.tmp$hapcat
iter=0;
for (m in cohort_list[-1]){
    #File.tmp=load(File);
    File.load=zaykin_all[[m]];
    cohort=paste("cohort", m, sep="");
    File.prep<-meta_hap_prep(File.load, cohort, annote=TRUE);
    if (sum(apply(File.prep$Hap.prob[, 1:(ncol(File.prep$Hap.prob)-2)], c(1,2), function(x){x=="TRUE"}))>0){
        stop ("genotype error!")
    }
    if (ncol(File.prep$Hap.prob)-2<N_variant) {
        merge.tmp=merge(merge.tmp, File.prep$Hap.prob[,-(ncol(File.prep$Hap.prob)-1)], by=names(File.prep$Hap.prob)[1:(ncol(File.prep$Hap.prob)-2)], all=TRUE);
    } else {
        merge.tmp=merge(merge.tmp, File.prep$Hap.prob[,c(1:N_variant, (N_variant+2))], by=lab, all=TRUE);
    }
    if (sum(is.na(merge.tmp$hapcat))==0){
        next
    }
    tmp=subset(merge.tmp, is.na(hapcat))
    tmp[order(eval(parse(text=paste("tmp$", cohort, ".hapcat", sep="")))), "hapcat"]=(max(merge.tmp[!is.na(merge.tmp$hapcat), "hapcat"])+1):(max(merge.tmp[!is.na(merge.tmp$hapcat), "hapcat"])+nrow(tmp))

    merge.tmp[is.na(merge.tmp$hapcat),"hapcat"]=tmp$hapcat;
    iter=iter+1;
}

N_W=max(merge.tmp$hapcat);
iter=0;
W=rep(0, N_W);
W.template=diag(1, N_W)
beta=NULL
SIGMA_i=0;

test.struct=rep(NA,2)
for (m in cohort_list){
    #File.tmp=load(File);
    File.load=zaykin_all[[m]];
    cohort=paste("cohort", m, sep="");
    File.prep<-meta_hap_prep(File.load, cohort, annote=TRUE);
    beta=c(beta, File.prep$Beta)
    print(File.prep$Beta)
    SIGMA_i=bdiag(SIGMA_i, solve(File.prep$Variance, tol=10^(-50)))
    merge.cohort=subset(merge.tmp, !is.na(eval(parse(text=paste(cohort, ".hapcat", sep="")))), select=c(names(merge.tmp)[1:(N_variant+1)], paste(cohort, ".hapcat", sep="")))
    merge.cohort=merge.cohort[order(eval(parse(text=paste("merge.cohort$", cohort, ".hapcat", sep="")))),]
    if (any(duplicated(merge.cohort[, ncol(merge.cohort)]))){
        merge.cohort=subset(merge.cohort, hapcat %in% tapply(merge.cohort$hapcat, eval(parse(text=paste(cohort, ".hapcat", sep=""))), function(x){min(x)}))
    }
    if (length(File.prep$Beta)<nrow(merge.cohort)){
        merge.cohort=subset(merge.cohort, eval(parse(text=paste(cohort, ".hapcat", sep=""))) %in% substr(names(File.prep$Beta), 4, nchar(File.prep$Beta)))
    } 
    level.comp=merge.cohort[eval(parse(text=paste("merge.cohort$", cohort, ".hapcat", sep="")))>1,"hapcat"] 
    level.ref=merge.cohort[eval(parse(text=paste("merge.cohort$", cohort, ".hapcat", sep="")))==1,"hapcat"] 
    test.struct=rbind(test.struct, cbind(level.comp, rep(level.ref, length(level.comp))))
    W.tmp=W.template[(merge.cohort$hapcat),]
    print(any(eigen(File.prep$Variance)$values==0))
    W=rbind(W, W.tmp)
    iter=iter+1;
}
test.struct= test.struct[-1,]
test.struct= t(apply(test.struct, 1, function(x){sort(x, decreasing=T)}))
test.struct=unique(test.struct)
test.struct=test.struct[order(test.struct[,2], decreasing=T),]

for (group in head(unique(test.struct[,2]),-1)){
if (group %in% test.struct[,1]){
test.struct[test.struct[,2]==group,2]=test.struct[test.struct[,1]==group,2][1]
test.struct=unique(test.struct)
}
}

test.struct2=test.struct[order(test.struct[,1]),]

W=W[-1,]
SIGMA_i=SIGMA_i[-1,-1]
SIGMA_i=as.matrix(SIGMA_i)
library(MASS)
beta_meta=solve(t(W)%*%SIGMA_i%*%W, tol=10^(-50))%*%t(W)%*%SIGMA_i%*%beta
beta_meta_var=solve(t(W)%*%SIGMA_i%*%W, tol=10^(-50))
beta_meta_trans=beta_meta[test.struct2[,1]]-beta_meta[test.struct2[,2]]
var_meta_trans=beta_meta_var[test.struct2[,1], test.struct2[,1]]+beta_meta_var[test.struct2[,2], test.struct2[,2]]-beta_meta_var[test.struct2[,1], test.struct2[,2]]-beta_meta_var[test.struct2[,2], test.struct2[,1]]
test_stats=matrix(beta_meta_trans, nrow=1) %*% solve(var_meta_trans, tol=10^(-50)) %*% matrix(beta_meta_trans, ncol=1)
#p-value of global test of independence between haplotypes and trait:
p.overall=pchisq(test_stats, df=length(beta_meta_trans), lower.tail=FALSE)
#p-value of individual test of any single haplotype
p.ind=pnorm(abs(beta_meta_trans)/sqrt(diag(var_meta_trans)), lower.tail=FALSE)*2

